/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef LXML_H
#define LXML_H

/* #ifdef USE_MMAP */
#include <sys/stat.h>
/* #endif */

#include "cntree.h"
#include "clist.h"

typedef struct LXMLhandle_s LXMLhandle;

typedef struct LXMLelement_s LXMLelement;

typedef struct LXMLstore_s LXMLstore;

typedef struct LXMLelement_callbacks_s LXMLelement_callbacks;

/*! Structure mapping an attribute to application specific data.  
 * This structure is used to convert data between the XML attributes
 * of an element and the application specific data.  Each XML
 * attribute has a type 'value_type' (string, int,
 * boolean). Associated to it is 'dest', a pointer to a pointer 
 * pointing to where the application specific data is stored.  When
 * converting from tree structure to application specific data, the
 * value of the attribute is converted from its original string format
 * to its 'value_type' (int, boolean. No conversion done for string
 * type).  This value is then copied into the application data at '*dest'. 
 * If the dest_kind is of kind LXML_Unique then an error is issued if 
 * there were already a data (check possible only for string value_type).
 * If the dest_kind is of kind LXML_List, then the value is appended to
 * the list pointed to by *dest
 * When converting from application data to tree structure,
 * then if 'dest' and '*dest' are not NULL, an attribute (or a list of
 * attributes) are created from the values in '*dest' after conversion
 * from their 'value_type' to a string.
 */

struct LXMLstore_s
{
  char *attr;      /**< the name of the attribute */
  int value_type;  /**< the type of attribute's value: LXML_String, LXML_Int 
		    *   or LXML_Boolean */
  void **dest;     /**< pointer to a pointer pointing to where the value is
		    *   (to be) stored. */
  int dest_kind;   /**< the kind of the destination '*dest': LXML_Unique, 
		    *   LXML_List */
};

/*! Structure containing the function callbacks used to convert between
 *  tree structure and application specific data.
 *  When converting from tree structure to application specific data, 
 *  for each element on the same depth  matching 'element_name', 
 *  the 'callback' function is called with the provided 'data'. 
 *  That callback is responsible to create the application specific data
 *  and link it to the supplied 'data'.
 *  When converting from application specific data to tree structure,
 *  whenever 'data' and '*data' are not NULL, a tree node (element of the XML
 *  file) is created with the corresponding 'element_name' and 'callback' is
 *  called with the provided 'data' argument and '*data_iter' initially set to NULL. 
 *  The 'callback' is called again and again (with a newly created element as its first 
 *  argument) as long as it returns 0 and '*data_iter' is non NULL. This allows 
 *  the 'callback' function to iterate through application specific 
 *  data using the '*data_iter' iterator that it uses at its convenience.
 *   When 'callback' does not need to be called again, it sets
 *  '*data_iter' to NULL, meaning that at this depth, no more elements 'element_name'
 *  are to be created. After each successful 'callback' return, the created element
 *  is added to the list of children of its parent element.
 */
struct LXMLelement_callbacks_s
{
  char *element_name; /**< The name of the element */
  int (*callback)(cntree *atree, void **data, void **data_iter, int op); /**< Callback function */
  void **data; /**< User provided data passed to the callback function */
};

enum
  {
    LXML_String = 1, /**< Attribute value is a string */
    LXML_Int,        /**< Attribute value is an integer */
    LXML_Boolean     /**< Attribute value is a boolean ('true', 'yes', 'false', 'no') */
  };

enum
  {
    LXML_Unique = 1, /**< Attribute value must be unique */
    LXML_List        /**< Mutlitple attributes with the same name are possible,
		      *   and application specific data is a list 
		      */
  };

enum
  {
    LXML_None, /**< Invalid file mode, used in the unintialized state */
    LXML_Read, /**< Open file in read mode, conversion from XML -> application
		*   specific data */
    LXML_Write /**< Openfile in write mode, conversion from application specific
		*   data to XML */
  };

/* XML file <-> tree conversion */

/*! Create a new handle to access an XML file 
 *  \return a newly allocated LXMLhandle to be freeed with lxml_free()
 */
LXMLhandle *
lxml_new(void);

/*! Free the ressources used by the handle. If lxml_file_open()
 *  was called before, you need to call lxml_file_close() yourself
 *  \param handle the handle created by lxml_new()
 */
void
lxml_free(LXMLhandle *handle);

/*! Set the name of the XML file
 *  \param handle the LXML handle
 *  \param filename the name of the file
 *  \return 0 on success, -1 on memory error
 */
int
lxml_set_filename(LXMLhandle *handle, const char *filename);

/*! Set the encoding of the XML file
 *  This is only used in the prolog of the file. No conversion is
 *  done by LXML. Strings are written as they are. FIXME: quotes
 *  should be escaped.
 *  \param handle the LXML handle
 *  \param encoding the encoding to be used in the XML prologue
 *  \return 0 on success, -1 on memory error
 */
int
lxml_set_encoding(LXMLhandle *handle, const char *encoding);

/*! Return the encoding as read from the file prologue
 *  The file must be read with lxml_file_read() before
 *  using this function. FIXME: not implemented yet
 *  \param handle the LXML handle
 *  \return the encoding of the file, as read by lxml_file_read() or
 *          as set by lxml_set_encoding. The former case is not implemented
 *          yet.
 */
const char *
lxml_get_encoding(LXMLhandle *handle);

/*! Set the DTD of the XML file
 *  This is only used in the prolog of the file. 
 *  This corresponds to a SYSTEM DTD.
 *  \param handle the LXML handle
 *  \param dtd the DTD to be used in the XML prologue
 *  \return 0 on success, -1 on memory error
 */
int
lxml_set_dtd(LXMLhandle *handle, const char *dtd);

/*! Return the DTD as read from the file prologue
 *  The file must be read with lxml_file_read() before
 *  using this function. FIXME: not implemented yet
 *  \param handle the LXML handle
 *  \return the DTD of the file, as read by lxml_file_read() or
 *          as set by lxml_set_dtd. The former case is not implemented
 *          yet.
 */
const char *
lxml_get_dtd(LXMLhandle *handle);

/*! Open the XML file for reading or writing
 *  The file name must have been set with lxml_set_filename() before
 *  calling this function
 *  \param handle the LXML handle
 *  \param mode LXML_Read or LXML_Write
 *  \return 0 on success or -1 on error.
 */
int
lxml_file_open(LXMLhandle *handle, int mode);
  
/*! Close the XML file
 *  If the file was opened in LXML_Read mode, it is munmap'ed.
 *  \param handle the LXML handle
 *  \return 0 on success or -1 on error.
 */
int
lxml_file_close(LXMLhandle *handle);

/*! Read the XML file and convert it to a tree structure
 *  The tree structure consist of a tree of nodes. Each
 *  node represents an element of the XML file.
 *  The tree is a cntree (see cntree.h). Each node
 *  points to the associated element and contains
 *  a list of its children along with a pointer to its parent.
 *  The element is of type LXMLElement and contains the
 *  name of the element, a flag to inform whether it has
 *  children elements and a list of its attributes.
 *  \param handle the LXML handle representing the file to read.
 *         Before using lxml_file_read(), lxml_file_open() must
 *         have been called with the LXML_Read mode
 *  \return the cntree representing the XML tree, or NULL on error.
 */
cntree *
lxml_file_read(LXMLhandle *handle);

/*! Write an XML tree representing the given tree
 *  If the encoding was previously set with lxml_set_encoding(), it
 *  will be written with the prologue of the XML file.
 *  If the DTD was previously set with lxml_set_dtd(), then
 *  the document type declaration will be written.
 *  \param handle the LXML handle representing the file to write.
 *         Before using lxml_file_write(), lxml_file_open() must
 *         have been called with the LXML_Write mode
 *  \param ntree the cntree representing an XML tree. The data of each
 *         node must be a pointer to an LXMLelement.
 *  \return 0 on success and -1 on error.
 */ 
int
lxml_file_write(LXMLhandle *handle, cntree *ntree);

/* tree handling functions */

/*! Create an empty tree with a single unnamed root element
 *  \return a newly allocated cntree which node is a pointer to
 *          an unname root element (LXMLelement)
 */
cntree *
lxml_tree_new(void);

/*! Free an XML tree
 *  Every resources used by the tree is freeed.
 *  \param ntree a cntree representing an XML tree.  Each node
 *         of the tree must be a pointer to an LXMLelement.
 */
void
lxml_tree_free(cntree *ntree);

/*! Free the XML tree but do not the attribute values
 *  Currently not used. Deprecated.
 *  \param ntree a cntree representing an XML tree.  Each node
 *         of the tree must be a pointer to an LXMLelement.
 */
void
lxml_tree_free_structural_data(cntree *ntree);

/*! Get the element associated to the current tree 
 *  \param ntree a cntree representing an XML tree.  Each node
 *         of the tree must be a pointer to an LXMLelement.
 *  \return the element pointed by the tree node ntree
 */
LXMLelement *
lxml_tree_get_element(cntree *ntree);

/* tree <-> application data conversion */

/*! For each node referenced by the given path, call a user provided
 *  callback.
 *  This is currently not used.  This functions traverses the XML
 *  tree and for each path matching the given path, it calls the user
 *  provided callback.
 *  \param ntree a cntree representing an XML tree.  Each node
 *         of the tree must be a pointer to an LXMLelement.
 *  \param path a NULL terminated array of element names representing
 *         the path to be matched
 *  \param callback the user callback to call for each matching element
 *  \param data the user provided data to pass to the callback function
 *  \param op the conversion direction (LXML_Read or LXML_Write)
 *  \return 0 on success, -1 on error.
 */
int 
lxml_tree_for_each_element_in_path(cntree *ntree, char **path,
			       int (*callback)(cntree *atree, 
					       void *adata, 
					       int aop), 
			       void *data,
			       int op);

/*! Convert between an XML element and application specific data.
 *  This function is to be called when convertin from or to an 
 *  XML tree, depending on the 'op' flag (resp. LXML_Read or 
 *  LXML_Write).
 *  This is the main function of LXML. It is passed a structure
 *  containing pointers to callback functions responsible for the
 *  actual conversion.
 *  When converting from an XML tree (LXML_Read), after an XML file
 *  was read, the callbacks functions listed in 'element_cb' are
 *  called whenever a corresponding element is matched. The callback
 *  function is called with the following arguments: the associated tree
 *  node, an associated, a user provided pointer to a pointer, 
 *  a pointer to NULL and the LXML_Read flag. The callback is responsible
 *  to create the application data associated to the element and to initialize
 *  it, eventually by using the attributes of the element.
 *  When converting to an XML tree, an initial root element (tree) must
 *  be provided by the user. That element must have its name set to the
 *  root element of the file.  Then for each 'element_cb' data that is non NULL
 *  or that points to a non-NULL pointer, the associated user callback is
 *  called with the following arguments: a newly created tree node with
 *  the associated element name, the associated LXMLelement_callbacks::data,
 *  a pointer to a pointer pointing initially to NULL and meant to be
 *  used by the callback function as an iterator, and the LXML_Write flag.
 *  The 'data_iter' is initially set to NULL for each element. The
 *  callback function can therefore detect that it was called for the 
 *  first time. If the application specific data is not a single data but
 *  a container, the callback function sets the pointer pointed to be 'data_iter' to
 *  the start iterator associated with the container. It then processes
 *  the data pointed to by the iterator, then it advances the iterator. If
 *  the iterator reaches the end of the container, it must be set to NULL.
 *  This is because LXML will continously call the same callback when it detects
 *  that the pointer pointed to by data_iter is non NULL, which means that there
 *  are still nodes to be created with the same element at that same depth.
 *  \param ntree a cntree representing an XML tree. The element name must be already
 *         set
 *  \param element_cb an array of LXMLelement_callbacks used to convert between
 *         XML elements and application specific data
 *  \param op a flag representing the direction of the conversion: LXML_Read
 *         when converting from an XML tree, and LXML_Write when converting
 *         to an XML tree
 *  \return 0 on success and -1 on failure
 */
int
lxml_tree_convert(cntree *ntree, LXMLelement_callbacks *element_cb, int op);

/* Element handling functions */

/*! Check or set the name of an element
 *  \param element the element which name is to be checked or set
 *  \param name the name to compare with or to set
 *  \param op LXML_Read to check if the element has the given name,
 *         or LXML_Write to set the name of the element to the given
 *         name
 *  \return 0 on success (element name matches or successfully set), -1
 *          on error (element name did not match or not set successfully).
 */
int
lxml_element_name(LXMLelement *element, char *name, int op);

/*! Store the attributes of an element
 *  Together with lxml_convert_tree(), this is second most important/used
 *  function in LXML.  It is used to convert between the XML attributes
 *  of an element and the application specific data representation.
 *  For this task, it uses the LXMLstore structure that describes
 *  the attributes and where their values are stored in the 
 *  application specific data representation.
 *  This function is typically used from callbacks called by 
 *  lxml_tree_convert().
 *  When converting from an XML tree (op flag set to LXML_Read),
 *  we need to use the attributes  values of an element to store them into
 *  the application specific data representation. In this case,
 *  this function will check the attributes of the element and if it 
 *  finds an attribute matching one listed in the 'store' LXMLstore
 *  array, it convert it according to the LXMLstore::value_type
 *  and store it in the associated pointer LXMLstore::dest.
 *  If the LXMLstore::dest_type is of type LXML_List, then
 *  the value is stored in the list pointed to by LXMLstore::dest.
 *  When converting to an XML tree (op flag set to LXML_Write),
 *  this function iterates through the array 'store', and
 *  if LXMLstore::dest is not NULL and if *LXMLstore::dest is also
 *  not NULL, then a new attribute is created with the value
 *  of *(LXMLstore::dest) converted to string. The new attribute
 *  is then appended to the list of attributes of the element. If
 *  the LXMLstore::dest_type is of type LXML_List, then the list
 *  traversed and for each data a new attribute is created and
 *  appended.
 *  \param element the element that contain the attributes
 *  \param store a NULL terminated array of LXMLstore describing
 *         the mapping between attributes and application specific
 *         data representation
 *  \param op the conversion direction flag. LXML_Read to convert
 *         from attributes, and LXML_Write to convert to attributes
 *  \return 0 on success and -1 on error
 */
int
lxml_element_store_attr(LXMLelement *element, LXMLstore *store, int op);

/*! Return the number of attributes of the given element
 *  \param element the element which attributes are to be counted
 *  \return the number of attributes of the element
 */
int
lxml_element_attr_count(LXMLelement *element);

#endif

/* vim: set cindent shiftwidth=2 tabstop=2 expandtab :*/  

