/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include "x_alloc.h"
#include "logging.h"

#include "ui.h"

#include "manager.h"
#include "folder.h"
#include "message.h"

#define XEP_MANAGER_MSG_LEN (sizeof(void *))

/* TODO: The number of maximal concurrent threads  should be user definable */
#define XEP_MANAGER_MAX_THREADS 10 

typedef struct XePManagerCallback_s XePManagerCallback;

struct XePManagerCallback_s
{
  XePManagerMsgCb callback_func;
  void *data;
};

static XePManagerThread *
Manager_thread_new(void);

static void
Manager_thread_init(XePManagerThread *thread, int write_fd);

static void
Manager_thread_free(XePManagerThread *thread);

static void
Manager_thread_reset(XePManagerThread *thread);

static void
Manager_thread_set_state(XePManagerThread *thread,
			 XePManagerThreadState state);

static XePManagerThreadState
Manager_thread_get_state(XePManagerThread *thread);

static void
Manager_thread_set_cancel(XePManagerThread *thread);

static void
Manager_thread_unset_cancel(XePManagerThread *thread);

static int
Manager_thread_get_cancel(XePManagerThread *thread);

static int
Manager_thread_queue_isempty(XePManagerThread *thread);

static void
Manager_thread_add_msg(XePManagerThread *thread, XePManagerMsg *msg);

static XePManagerMsg *
Manager_thread_get_msg(XePManagerThread *thread);

static int
Manager_thread_create(XePManagerThread *thread);

static void
Manager_thread_signal(XePManagerThread *thread);

static int
Manager_thread_queue_lock(XePManagerThread *thread);

static int
Manager_thread_queue_unlock(XePManagerThread *thread);

static int
Manager_thread_data_lock(XePManagerThread *thread);

static int
Manager_thread_data_unlock(XePManagerThread *thread);

static void *
Manager_thread(void *athread);

static void
Manager_thread_execute_msg(XePManagerMsg *msg, XePManagerThread *thread);

static void
Manager_queue_add_msg(clist *queue, XePManagerMsg *msg);

static XePManagerMsg *
Manager_queue_get_msg(clist *queue);

static void
Manager_pending_queue_free(void);

static int
Manager_pending_queue_isempty(void);

static void
Manager_pending_queue_add_msg(XePManagerMsg *msg);

static XePManagerMsg *
Manager_pending_queue_get_msg(void);

static clist *
Manager_pending_queue_get_other_queue(void *target);

static void
Manager_pending_queue_concat(XePManagerThread *thread);

static void
Manager_queue_add_msg(clist *queue, XePManagerMsg *msg);

static XePManagerMsg *
Manager_queue_get_msg(clist *queue);

static void
Manager_dump_queue(clist *queue);

static void
Manager_dump_msg(XePManagerMsg *msg);

static void
Manager_manage_incoming(int fd, short int condition, void *data);

static void
Manager_manage_execute(XePManagerMsg *msg);

static void
Manager_manage_done(XePManagerMsg *msg);

static void
Manager_manage_exit(XePManagerMsg *msg);

static void
Manager_submit_msg(int fd, XePManagerMsg *msg);

static void
Manager_submit_queue(int fd, clist *queue);

/* Static variables */
static pthread_mutex_t manager_mutex = PTHREAD_MUTEX_INITIALIZER;

static XePManager *global_manager;


XePManagerThread *
Manager_thread_new(void)
{
  XePManagerThread *thread;

  thread = x_alloc(1, XePManagerThread);
  if (thread == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }
  else
    {
      Manager_thread_init(thread, 0);
    }
  return thread;
}

void
Manager_thread_init(XePManagerThread *thread, int fd_write)
{
  pthread_mutex_init(&thread->queue_mutex, NULL);
  pthread_mutex_init(&thread->data_mutex, NULL);
  pthread_cond_init(&thread->cond, NULL);
  thread->state = XEP_MANAGER_THREAD_STATE_NONE;
  thread->queue = clist_new();
  thread->fd_write = fd_write;
  thread->cancel = 0;
}

void
Manager_thread_free(XePManagerThread *thread)
{
  x_free(thread);
}


static int
Manager_thread_create(XePManagerThread *thread)
{
  int r;
  pthread_attr_t attr;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  Manager_thread_init(thread,  global_manager->fd_write);

  r = pthread_create(&thread->thread_id,
		     NULL,
		     Manager_thread,
		     (void *) thread);

  pthread_attr_destroy(&attr);
  
  if (r != 0) 
    {
      A_LOGGING_SYSTEM_ERROR("pthread_create: %s", strerror(r));
    }
  else
    {
      Manager_thread_set_state(thread, XEP_MANAGER_THREAD_STATE_RUNNING);
      global_manager->thread_count++;
#if DEBUG
      printf("New thread created with id %ld. Nb of threads = %d\n", 
	     thread->thread_id, global_manager->thread_count);
#endif
    }

  return r;
}

static void
Manager_thread_signal(XePManagerThread *thread)
{
  pthread_cond_signal(&thread->cond);
}

/** destroy mutex and condition and set thread state to none
 */
void
Manager_thread_reset(XePManagerThread *thread)
{
  
  if (pthread_mutex_destroy(&thread->queue_mutex) == EBUSY)
    {
      A_LOGGING_SYSTEM_ERROR("pthread_mutex_destroy: %s", strerror(errno));
    }
  else if (pthread_mutex_destroy(&thread->data_mutex) == EBUSY)
    {
      A_LOGGING_SYSTEM_ERROR("pthread_mutex_destroy: %s", strerror(errno));
    }
  else if (pthread_cond_destroy(&thread->cond) == EBUSY)
    {
      A_LOGGING_SYSTEM_ERROR("pthread_cond_destroy: %s", strerror(errno));
    }
  else 
    {
      Manager_thread_set_state(thread, XEP_MANAGER_THREAD_STATE_NONE);
      global_manager->thread_count--;
    }
}

static void
Manager_thread_set_state(XePManagerThread *thread,
			 XePManagerThreadState state)
{
  thread->state = state;
}

static XePManagerThreadState
Manager_thread_get_state(XePManagerThread *thread)
{
  return thread->state;
}

XePManagerThreadState
a_Manager_thread_get_state_safe(XePManagerThread *thread)
{
  XePManagerThreadState state;
  Manager_thread_queue_lock(thread);
  state = Manager_thread_get_state(thread);
  Manager_thread_queue_unlock(thread);

  return state;
}

static void
Manager_thread_set_cancel(XePManagerThread *thread)
{
  thread->cancel = 1;
}

static void
Manager_thread_unset_cancel(XePManagerThread *thread)
{
  thread->cancel = 0;
}

static int
Manager_thread_get_cancel(XePManagerThread *thread)
{
  return thread->cancel;
}

static int
Manager_thread_queue_isempty(XePManagerThread *thread)
{
  return clist_isempty(thread->queue);
}

static int
Manager_thread_queue_lock(XePManagerThread *thread)
{
  int r;
  A_LOGGING_TRACE;

  r = pthread_mutex_lock(&thread->queue_mutex);
  if (r)
    {
      A_LOGGING_SYSTEM_ERROR("pthread_mutex_lock: %s", strerror(r));
    }
  return r;
}

static int
Manager_thread_queue_unlock(XePManagerThread *thread)
{
  int r;
  A_LOGGING_TRACE;

  r = pthread_mutex_unlock(&thread->queue_mutex);
  if (r)
    {
      A_LOGGING_SYSTEM_ERROR("pthread_mutex_unlock: %s", strerror(r));
    }
  return r;
}

static void
Manager_thread_queue_wait(XePManagerThread *thread)
{
  A_LOGGING_TRACE;

  pthread_cond_wait(&thread->cond, &thread->queue_mutex);
}


static int
Manager_thread_data_lock(XePManagerThread *thread)
{
  int r;
  A_LOGGING_TRACE;

  r = pthread_mutex_lock(&thread->data_mutex);
  if (r)
    {
      A_LOGGING_SYSTEM_ERROR("pthread_mutex_lock: %s", strerror(r));
    }
  return r;
}

static int
Manager_thread_data_unlock(XePManagerThread *thread)
{
  int r;
  A_LOGGING_TRACE;

  r = pthread_mutex_unlock(&thread->data_mutex);
  if (r)
    {
      A_LOGGING_SYSTEM_ERROR("pthread_mutex_unlock: %s", strerror(r));
    }
  return r;
}
void
Manager_thread_add_msg(XePManagerThread *thread, XePManagerMsg *msg)
{
  Manager_queue_add_msg(thread->queue, msg);
}

XePManagerMsg *
Manager_thread_get_msg(XePManagerThread *thread)
{
  return Manager_queue_get_msg(thread->queue);  
}

static XePManagerCallback *
Manager_callback_new(XePManagerMsgCb callback_func, void *data)
{
  XePManagerCallback *callback;
  callback = x_alloc(1, XePManagerCallback);
  if (callback)
    {
      callback->callback_func = callback_func;
      callback->data = data;
    }
  else
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }
  return callback;
}

static void
Manager_callback_free(XePManagerCallback *callback)
{
  x_free(callback);
}

XePManagerMsg *
a_Manager_msg_new(void)
{
  XePManagerMsg *msg;
  
  msg = x_alloc(1, XePManagerMsg);

  if (msg == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }
  msg->source = NULL;
  msg->target = NULL;
  msg->arg_in = NULL;
  msg->arg_out = NULL;
  msg->callbacks = NULL;
  msg->state = NULL;
  msg->flush_old = 0;
  return msg;
}

void
a_Manager_msg_free(XePManagerMsg *msg)
{
  if (msg)
    {
      x_free(msg);
    }
}

void
a_Manager_msg_set_type(XePManagerMsg *msg, XePManagerMsgType type)
{
  msg->type = type;
}

XePManagerMsgType
a_Manager_msg_get_type(XePManagerMsg *msg)
{
  return msg->type;
}

void
a_Manager_msg_flush_old(XePManagerMsg *msg)
{
  msg->flush_old = 1;
}

void
a_Manager_msg_set_source(XePManagerMsg *msg, void *source)
{
  msg->source = source;
}

void *
a_Manager_msg_get_source(XePManagerMsg *msg)
{
  return msg->source;
}


void
a_Manager_msg_set_target(XePManagerMsg *msg, void *target)
{
  msg->target = target;
}

void *
a_Manager_msg_get_target(XePManagerMsg *msg)
{
  return msg->target;
}


void
a_Manager_msg_set_source_target(XePManagerMsg *msg, 
				void *source, 
				void *target)
{
  msg->source = source;
  msg->target = target;
}

void
a_Manager_msg_add_callback(XePManagerMsg *msg,
			   XePManagerMsgCb callback_func,
			   void *data)
{
  if (msg) 
    {
      if (msg->callbacks == NULL)
	{
	  msg->callbacks = clist_new();
	}
      if (msg->callbacks)
	{
	  XePManagerCallback *callback = Manager_callback_new(callback_func,
							      data);
	  clist_append(msg->callbacks, callback);
	}
    }
  else
    {
      A_LOGGING_SYSTEM_ERROR("msg is NULL");
    }
}

void
a_Manager_msg_set_arg_in(XePManagerMsg *msg, void *arg_in)
{
  msg->arg_in = arg_in;
}

void *
a_Manager_msg_get_arg_in(XePManagerMsg *msg)
{
  return msg->arg_in;
}

void
a_Manager_msg_set_arg_out(XePManagerMsg *msg, void *arg_out)
{
  msg->arg_out = arg_out;
}

void *
a_Manager_msg_get_arg_out(XePManagerMsg *msg)
{
  return msg->arg_out;
}

void
a_Manager_msg_set_args(XePManagerMsg *msg, void *arg_in, void *arg_out)
{
  msg->arg_in  = arg_in;
  msg->arg_out = arg_out;
}

void
a_Manager_msg_set_state_holder(XePManagerMsg *msg,
			       XePManagerMsgType *state_holder)
{
  msg->state = state_holder;
  *(msg->state) = msg->type;
}

/* This function is called from any thread */
void
a_Manager_msg_submit(XePManagerMsg *msg)
{
  Manager_submit_msg(global_manager->fd_write, msg);
}

static void
Manager_pending_queue_free(void)
{
  clistiter *iter;
  for (iter = clist_begin(global_manager->pending_queue); 
       iter; 
       iter = iter->next)
    {
      a_Manager_msg_free((XePManagerMsg *) iter->data);
    }
  clist_free(global_manager->pending_queue);
}

static int
Manager_pending_queue_isempty(void)
{
  return clist_isempty(global_manager->pending_queue);
}

static void
Manager_pending_queue_add_msg(XePManagerMsg *msg)
{
  Manager_queue_add_msg(global_manager->pending_queue, msg);
}

static XePManagerMsg *
Manager_pending_queue_get_msg(void)
{
  return Manager_queue_get_msg(global_manager->pending_queue);
}

/** Manager_pending_queue_get_other_queue
 *  Return the first msg in the queue that is not to be handled by TARGET.
 *  If there is no such msg, the first msg to be handled by target is returned
 *  TODO: it should return a list of correspondig msg and not only one... 
 */
static clist *
Manager_pending_queue_get_other_queue(void *target)
{
  clistiter *iter;
  clist *other_queue = NULL;
  XePManagerMsg *msg = NULL;

  if (!Manager_pending_queue_isempty())
  {
    other_queue = clist_new();

    for (iter = clist_begin(global_manager->pending_queue); 
	 iter; 
	 iter = iter->next)
      {
	msg = (XePManagerMsg *) iter->data;
	if (msg->target != target)
	  {
	  target = msg->target;
	  iter = NULL;
	  }
      }
    for (iter = clist_begin(global_manager->pending_queue); 
	 iter; 
	 iter = iter->next)
      {
	msg = (XePManagerMsg *) iter->data;
	if (msg->target == target)
	  {
	    clistiter *next = iter->next;
	    clist_append(other_queue, msg);
	    clist_delete(global_manager->pending_queue, iter);
	    iter = next;
	  }
      }
  }

  return other_queue;
}

static void
Manager_pending_queue_concat(XePManagerThread *thread)
{
  clist_concat(global_manager->pending_queue, thread->queue);
}

static void
Manager_queue_add_msg(clist *queue, XePManagerMsg *msg)
{
  clist_append(queue, msg);
#if DEBUG
  Manager_dump_queue(queue);
#endif
}


static XePManagerMsg *
Manager_queue_get_msg(clist *queue)
{
  clistiter *iter;
  XePManagerMsg *msg = NULL;

  iter = clist_begin(queue);
  if (iter)
    {
      msg = (XePManagerMsg *) iter->data;
      clist_delete(queue, iter);
    }

  return msg;
}

static void
Manager_dump_queue(clist *queue)
{
  clistiter *iter;
  XePManagerMsg *msg;
  int i = 0;

  if (queue == NULL)
    printf("Queue NULL\n");
  else if (clist_isempty(queue))
    printf("Queue empty\n");
  else
    {
      printf("Queue %p\n", queue);
      for (iter = clist_begin(queue); iter; iter = iter->next)
	{
	  msg = (XePManagerMsg *) iter->data;
	  printf("Message %d\n", i++);
	  Manager_dump_msg(msg);
	}
    }
}

static void
Manager_dump_msg(XePManagerMsg *msg)
{
  if (msg == NULL)
    {
      printf("Manager_msg_dump(): msg is NULL");
    }
  else
    {
      printf("Msg %p Type: ", msg);
      switch (msg->type)
	{
	case XEP_MANAGER_CAPABILITY:
	  printf("XEP_MANAGER_CAPABILITY  ");
	  break;
	case XEP_MANAGER_CREATE:
	  printf("XEP_MANAGER_CREATE      ");
	  break;
	case XEP_MANAGER_DELETE:
	  printf("XEP_MANAGER_DELETE      ");
	  break;
	case XEP_MANAGER_RENAME:
	  printf("XEP_MANAGER_RENAME      ");
	  break;
	case XEP_MANAGER_SUBSCRIBE:
	  printf("XEP_MANAGER_SUBSCRIBE   ");
	  break;
	case XEP_MANAGER_UNSUBSCRIBE:
	  printf("XEP_MANAGER_UNSUBSCRIBE ");
	  break;
	case XEP_MANAGER_LIST:
	  printf("XEP_MANAGER_LIST        ");
	  break;
	case XEP_MANAGER_LSUB:
	  printf("XEP_MANAGER_LSUB        ");
	  break;
	case XEP_MANAGER_DISCONNECT:
	  printf("XEP_MANAGER_DISCONNECT  ");
	  break;
	case XEP_MANAGER_MAILBOX_DISCONNECT:
	  printf("XEP_MANAGER_MAILBOX_DISCONNECT");
	  break;
	case XEP_MANAGER_STATUS:
	  printf("XEP_MANAGER_STATUS      ");
	  break;
	case XEP_MANAGER_APPEND:
	  printf("XEP_MANAGER_APPEND      ");
	  break;
	case XEP_MANAGER_CLOSE:
	  printf("XEP_MANAGER_CLOSE       ");
	  break;
	case XEP_MANAGER_EXPUNGE:
	  printf("XEP_MANAGER_EXPUNGE     ");
	  break;
	case XEP_MANAGER_SEARCH:
	  printf("XEP_MANAGER_SEARCH      ");
	  break;
	case XEP_MANAGER_GET_BODYSTRUCTURE:
	  printf("XEP_MANAGER_GET_BODYSTRUCTURE");
	  break;
	case XEP_MANAGER_MIME_FIND_DEFAULT:
	  printf("XEP_MANAGER_MIME_FIND_DEFAULT");
	  break;
	case XEP_MANAGER_FETCH_FIELDS:
	  printf("XEP_MANAGER_FETCH_FIELDS  ");
	  break;
	case XEP_MANAGER_FETCH_PART:
	  printf("XEP_MANAGER_FETCH_PART  ");
	  break;
	case XEP_MANAGER_FETCH:
	  printf("XEP_MANAGER_FETCH       ");
	  break;
	case XEP_MANAGER_FLAGS_CHANGE:
	  printf("XEP_MANAGER_FLAGS_CHANGE   ");
	  break;
	case XEP_MANAGER_STORE:
	  printf("XEP_MANAGER_STORE       ");
	  break;
	case XEP_MANAGER_COPY:
	  printf("XEP_MANAGER_COPY        ");
	  break;
	case XEP_MANAGER_PROCESSING:
	  printf("XEP_MANAGER_PROCESSING  ");
	  break;
	case XEP_MANAGER_SEARCHING:
	  printf("XEP_MANAGER_SEARCHING   ");
	  break;
	case XEP_MANAGER_UPDATE:
	  printf("XEP_MANAGER_UPDATE      ");
	  break;
	case XEP_MANAGER_CHECK:
	  printf("XEP_MANAGER_CHECK      ");
	  break;
	case XEP_MANAGER_SMTP_SEND:
	  printf("XEP_MANAGER_SMTP_SEND  ");
	  break;
	case XEP_MANAGER_DONE:
	  printf("XEP_MANAGER_DONE        ");
	  break;
	case XEP_MANAGER_EXIT:
	  printf("XEP_MANAGER_EXIT        ");
	  break;
	case XEP_MANAGER_EXIT_DONE:
	  printf("XEP_MANAGER_EXIT_DONE   ");
	  break;
	default:
	  printf("Unknown!!!");
	}
      printf("\ns: %p ", msg->source);
      printf("t: %p ", msg->target);
      printf("i: %p ", msg->arg_in);
      printf("o: %p ", msg->arg_out);
      printf("c: %p ", msg->callbacks);
    }
}

void
a_Manager_init(XetPan *xetpan)
{
  int manager_pipe[2];

  A_LOGGING_DEBUG("<%ld>", pthread_self());

  global_manager = &xetpan->manager;

  if (pipe(manager_pipe))
    {
      A_LOGGING_SYSTEM_ERROR("pipe: %s", strerror(errno));
    }
  else
    {
      XePRemote *remote;

      remote = x_alloc(1, XePRemote);
      if (remote == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
	}
      else
	{
	  remote->fd = manager_pipe[0];
	  remote->xetpan = xetpan;
	  remote->msg = x_alloc(XEP_MANAGER_MSG_LEN, char);
	  remote->msg_size = XEP_MANAGER_MSG_LEN;
	  remote->msg_len = 0;
	  remote->msg_idx = 0;
	  remote->watch_id = a_UI_add_io_watch(xetpan, 
						manager_pipe[0], 
						XETPAN_POLLIN,
						Manager_manage_incoming,
						xetpan);
	  if (remote->watch_id == 0)
	    {
	      A_LOGGING_SYSTEM_ERROR("failed");
	      close(manager_pipe[0]);
	      close(manager_pipe[1]);
	      x_free(remote);
	    }
	  else 
	    {
	      global_manager->remote = remote;
	      global_manager->fd_write = manager_pipe[1];
	      global_manager->pending_queue = clist_new();
	      global_manager->max_threads = XEP_MANAGER_MAX_THREADS;
	      global_manager->thread_count = 0;
	    }
	}
    }
}

void
a_Manager_quit(XetPan *xetpan)
{
  a_UI_remove_io_watch(xetpan, global_manager->remote->watch_id);

  close(global_manager->remote->fd);
  close(global_manager->fd_write);

  Manager_pending_queue_free();
  x_free(global_manager->remote->msg);
  x_free(global_manager->remote);

  global_manager->fd_write = 0;
}

static void
Manager_manage_incoming(int fd, short int condition, void *data)
{
  XePRemote *remote = global_manager->remote;
  ssize_t count;

  A_LOGGING_DEBUG("<%ld>", pthread_self());

  if (remote->msg_idx < remote->msg_size)
    {
      count = read(fd, remote->msg + remote->msg_idx, 
		   remote->msg_size - remote->msg_idx);
      /* TODO: Better error handling code */
      if (count < 0)
	{
	  A_LOGGING_SYSTEM_ERROR("read: %s", strerror(errno));
	  return;
	}
      remote->msg_idx += count;
    }

  if (remote->msg_idx == remote->msg_size)
    {
      XePManagerMsg *msg;

      /* Reset remote msg buffer */
      remote->msg_idx = 0;

      msg = (XePManagerMsg*) (*(void **)remote->msg);
 
      switch (msg->type)
	{
	case XEP_MANAGER_DONE:
	  Manager_manage_done(msg);
	  break;
	case XEP_MANAGER_EXIT_DONE:
	  Manager_manage_exit(msg);
	  break;
	default:
	  Manager_manage_execute(msg);
	}
    }
}

static void
Manager_manage_execute(XePManagerMsg *msg)
{
  XePManagerThread *thread = msg->target;
  int r = 0;

  A_LOGGING_DEBUG("<%ld> msg %p", pthread_self(), msg);

  if (Manager_thread_get_state(thread) == XEP_MANAGER_THREAD_STATE_NONE)
    {
      if (global_manager->thread_count < global_manager->max_threads)
	{
	  r = Manager_thread_create(thread);
	}
      else
	{
	  r = EAGAIN;
	}
      if (r)
	{
	  A_LOGGING_DEBUG("maximum number of running threads already reached "
			  "submitting msg to pending queue");
	  Manager_pending_queue_add_msg(msg);
	}
    }

  if (Manager_thread_get_state(thread) != XEP_MANAGER_THREAD_STATE_NONE)
    {
      Manager_thread_queue_lock(thread);

      if (msg->flush_old)
	{
	  Manager_thread_set_cancel(thread);
	}
      Manager_thread_add_msg(thread, msg);

      Manager_thread_queue_unlock(thread);
      
      Manager_thread_signal(thread);
    }
}

static void
Manager_manage_done(XePManagerMsg *msg)
{
  void *arg_out = msg->arg_out;
  clist *callbacks = msg->callbacks;
  clistiter *iter;
  XePManagerMsgType *state_holder = NULL;

  A_LOGGING_DEBUG("<%ld> msg %p", pthread_self(), msg);

  Manager_dump_msg(msg);

  if (msg->state)
    {
      state_holder = msg->state;
    }

  a_Manager_msg_free(msg);

  if (callbacks)
    {
      /* TODO: maybe stop calling callbacks if one of the returns non zero */
      for (iter = clist_begin(callbacks); iter; iter = iter->next)
	{
	  XePManagerCallback *callback = (XePManagerCallback *) iter->data;
	  callback->callback_func(callback->data, arg_out);
	  Manager_callback_free(callback);
	}
      clist_free(callbacks);
    }

  if (state_holder)
    {
      *state_holder = XEP_MANAGER_DONE;
    }
}

static void
Manager_manage_exit(XePManagerMsg *msg)
{
  XePManagerThread *thread = msg->target;

  A_LOGGING_TRACE;

  Manager_dump_msg(msg);

  Manager_thread_queue_lock(thread);

  /* Add not handled thread msgs to the main pending queue for
   * later submission 
   */
  Manager_pending_queue_concat(thread);

  Manager_thread_queue_unlock(thread);

  Manager_thread_reset(thread);

  a_Manager_msg_free(msg);

  if (Manager_pending_queue_isempty())
    {
      clist *pending_queue;
      pending_queue = Manager_pending_queue_get_other_queue(thread);
      if (pending_queue)
	{
	  Manager_submit_queue(global_manager->fd_write, pending_queue);
	  clist_free(pending_queue);
	}
    }
}

static void *
Manager_thread(void *athread)
{
  XePManagerThread *thread = (XePManagerThread *) athread;
  XePManagerMsg *msg;
  int cancel;
  int loop = -1;
  int r;

  A_LOGGING_DEBUG("<%ld>", pthread_self());

  while (loop)
    {
      r = Manager_thread_queue_lock(thread);
      if (Manager_thread_queue_isempty(thread))
	{
	  Manager_thread_set_state(thread, XEP_MANAGER_THREAD_STATE_IDLE);
	  Manager_thread_queue_wait(thread);
	}

      Manager_thread_set_state(thread, XEP_MANAGER_THREAD_STATE_RUNNING);
     
      msg = Manager_thread_get_msg(thread);

      cancel = Manager_thread_get_cancel(thread);

      A_LOGGING_DEBUG("got msg=%p", msg);

      Manager_thread_queue_unlock(thread);

      if (msg->type == XEP_MANAGER_EXIT)
	{
	  loop = 0;
	  a_Manager_msg_set_type(msg, XEP_MANAGER_EXIT_DONE);
	}
      else
	{
	  if (cancel && msg->flush_old)
	    {
	      Manager_thread_queue_lock(thread);
	      Manager_thread_unset_cancel(thread);
	      cancel = 0;
	      Manager_thread_queue_unlock(thread);
	    }
	  if (!cancel)
	    {
	      Manager_thread_execute_msg(msg, thread);
	    }
	  else
	    {
	      A_LOGGING_DEBUG("msg %p cancelled", msg);
	    }
	  a_Manager_msg_set_type(msg, XEP_MANAGER_DONE);
	}

      Manager_submit_msg(thread->fd_write, msg);
    }

  /* FIXME: Testing... */
  A_LOGGING_DEBUG("ending thread thread_id=%ld", thread->thread_id);
  /* Just exit */
  return NULL;
}

static void
Manager_thread_execute_msg(XePManagerMsg *msg, XePManagerThread *thread)
{
  A_LOGGING_DEBUG("<%ld>", pthread_self());

  Manager_dump_msg(msg);
  
  switch (msg->type)
    {
    case XEP_MANAGER_CAPABILITY:
      A_LOGGING_DEBUG("XEP_MANAGER_CAPABILITY");
      break;
    case XEP_MANAGER_CREATE:
      A_LOGGING_DEBUG("XEP_MANAGER_CREATE");
      break;
    case XEP_MANAGER_DELETE:
      A_LOGGING_DEBUG("XEP_MANAGER_DELETE");
      break;
    case XEP_MANAGER_RENAME:
      A_LOGGING_DEBUG("XEP_MANAGER_RENAME");
      break;
    case XEP_MANAGER_SUBSCRIBE:
      A_LOGGING_DEBUG("XEP_MANAGER_SUBSCRIBE");
      break;
    case XEP_MANAGER_UNSUBSCRIBE:
      A_LOGGING_DEBUG("XEP_MANAGER_UNSUBSCRIBE");
      break;
    case XEP_MANAGER_LIST:
      A_LOGGING_DEBUG("XEP_MANAGER_LIST");
      Manager_thread_data_lock(thread);
      a_Folder_mailbox_list((XePFolder *) msg->arg_in, 
			    (struct mail_list **) msg->arg_out);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_LSUB:
      A_LOGGING_DEBUG("XEP_MANAGER_LSUB");
      Manager_thread_data_lock(thread);
      a_Folder_mailbox_lsub((XePFolder *) msg->arg_in, 
			    (struct mail_list **) msg->arg_out);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_DISCONNECT:
      A_LOGGING_DEBUG("XEP_MANAGER_DISCONNECT");
      Manager_thread_data_lock(thread);
      a_Folder_disconnect((XePFolder *) msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_MAILBOX_DISCONNECT:
      A_LOGGING_DEBUG("XEP_MANAGER_MAILBOX_DISCONNECT");
      Manager_thread_data_lock(thread);
      a_Folder_mailbox_disconnect((XePFolder *) msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_STATUS:
      A_LOGGING_DEBUG("XEP_MANAGER_STATUS");
      break;
    case XEP_MANAGER_APPEND:
      A_LOGGING_DEBUG("XEP_MANAGER_APPEND");
      Manager_thread_data_lock(thread);
      a_Message_append_to_folder((XePMessage *) msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_CLOSE:
      A_LOGGING_DEBUG("XEP_MANAGER_CLOSE");
      break;
    case XEP_MANAGER_EXPUNGE:
      A_LOGGING_DEBUG("XEP_MANAGER_EXPUNGE");
      Manager_thread_data_lock(thread);
      a_Folder_expunge((XePFolder *) msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_SEARCH:
      A_LOGGING_DEBUG("XEP_MANAGER_SEARCH");
      break;
    case XEP_MANAGER_GET_BODYSTRUCTURE:
      A_LOGGING_DEBUG("XEP_MANAGER_GET_BODYSTRUCTURE");
      Manager_thread_data_lock(thread);
      a_Message_get_bodystructure(msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_MIME_FIND_DEFAULT:
      A_LOGGING_DEBUG("XEP_MANAGER_MIME_FIND_DEFAULT");
      Manager_thread_data_lock(thread);
      a_Message_mime_find_default(msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_FETCH_PART:
      A_LOGGING_DEBUG("XEP_MANAGER_FETCH_PART");
      Manager_thread_data_lock(thread);
      a_Message_fetch_part(msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;      
    case XEP_MANAGER_FETCH:
      A_LOGGING_DEBUG("XEP_MANAGER_FETCH <%ld>", pthread_self());
      Manager_thread_data_lock(thread);
      a_Message_fetch(msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_FLAGS_CHANGE:
      A_LOGGING_DEBUG("XEP_MANAGER_FLAGS_CHANGE");
      Manager_thread_data_lock(thread);
      a_Message_flags_change(msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_STORE:
      A_LOGGING_DEBUG("XEP_MANAGER_STORE");
      break;
    case XEP_MANAGER_COPY:
      A_LOGGING_DEBUG("XEP_MANAGER_COPY");
      break;
    case XEP_MANAGER_PROCESSING:
      A_LOGGING_DEBUG("XEP_MANAGER_PROCESSING");
      break;
    case XEP_MANAGER_SEARCHING:
      A_LOGGING_DEBUG("XEP_MANAGER_SEARCHING");
      break;
    case XEP_MANAGER_UPDATE:
      A_LOGGING_DEBUG("XEP_MANAGER_UPDATE <%ld>", pthread_self());
      Manager_thread_data_lock(thread);
      a_Folder_sync(msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_CHECK:
      A_LOGGING_DEBUG("XEP_MANAGER_CHECK <%ld>", pthread_self());
      Manager_thread_data_lock(thread);
      a_Folder_check(msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    case XEP_MANAGER_SMTP_SEND:
      A_LOGGING_DEBUG("XEP_MANAGER_SMTP_SEND <%ld>", pthread_self());
      Manager_thread_data_lock(thread);
      msg->arg_out = (void *) a_MTA_smtp_send_message(msg->arg_in);
      Manager_thread_data_unlock(thread);
      break;
    default:
      A_LOGGING_SYSTEM_ERROR("unknown msg type %d", msg->type);
    }
}

/* This function is called from any thread */
static void
Manager_submit_msg(int fd, XePManagerMsg *msg)
{
  char *msg_idx = (char *) &msg;
  ssize_t count = 0;
  ssize_t sent_count = 0;

  A_LOGGING_DEBUG("<%ld> submitting msg=%p", pthread_self(), msg);

  pthread_mutex_lock(&manager_mutex);
  while (sent_count < XEP_MANAGER_MSG_LEN)
    {
      count = write(fd, 
		    msg_idx + sent_count, 
		    XEP_MANAGER_MSG_LEN - sent_count);
      if (count < 0)
	{
	  A_LOGGING_SYSTEM_ERROR("write: %s", strerror(errno));
	}
      else 
	sent_count += count;
    }
  pthread_mutex_unlock(&manager_mutex);

}

static void
Manager_submit_queue(int fd, clist *queue)
{
  clistiter *iter;
  XePManagerMsg *msg;
  A_LOGGING_TRACE;
  for (iter = clist_begin(queue); iter; iter = iter->next)
    {
      msg = (XePManagerMsg *) iter->data;
      Manager_submit_msg(fd, msg);
    }
}
