/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include <clist.h>


#include "intl.h"

#include "logging.h"
#include "x_alloc.h"

#include "lxml.h"
#include "processing.h"

static int
Processing_conditions_cb(cntree *atree, void **data, void **data_iter, int op);

static int
Processing_actions_cb(cntree *atree, void **data, void **data_iter, int op);

static int
Processing_rules_cb(cntree *atree, void **data, void **data_iter, int op);

static int
Processing_condition_cb(cntree *atree, void **data, void **data_iter, int op);

static int
Processing_action_cb(cntree *atree, void **data, void **data_iter, int op);

static int
Processing_rule_cb(cntree *atree, void **data, void **data_iter, int op);

XePProcessing *
a_Processing_new(void)
{
  chash *rules;
  chash *conditions;
  chash *actions;
  XePProcessing *processing;

  rules = chash_new(CHASH_DEFAULTSIZE, CHASH_COPYNONE);

  if (rules == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("chash_new returned NULL");
      goto err_rules;
    }

  conditions = chash_new(CHASH_DEFAULTSIZE, CHASH_COPYNONE);

  if (conditions == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("chash_new returned NULL");
      goto err_conditions;
    }

  actions = chash_new(CHASH_DEFAULTSIZE, CHASH_COPYNONE);

  if (actions == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("chash_new returned NULL");
      goto err_actions;
    }

  processing = x_alloc(1, XePProcessing);

  if (processing == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      goto err_processing;
    }

  processing->rules      = rules;
  processing->conditions = conditions;
  processing->actions    = actions;

  return processing;

 err_processing:
  chash_free(actions);
 err_actions:
  chash_free(conditions);
 err_conditions:
  chash_free(rules);
 err_rules:
  return NULL;
}

void
a_Processing_free(XePProcessing *processing)
{
  chashiter  *hiter;
  chashdatum hiter_datum;
  chash *hash;

  if (processing == NULL)
    return;

  if ((hash = processing->rules))
    {
      for (hiter = chash_begin(hash); hiter; hiter = chash_next(hash, hiter))
	{
	  chash_value(hiter, &hiter_datum);
	  if (hiter_datum.data)
	    a_Processing_rule_free((XePProcessingRule *)hiter_datum.data);
	}
      chash_free(hash);
    }

  if ((hash = processing->conditions))
    {
      for (hiter = chash_begin(hash); hiter; hiter = chash_next(hash, hiter))
	{
	  chash_value(hiter, &hiter_datum);
	  if(hiter_datum.data)
	    a_Processing_condition_free((XePProcessingCondition *)
					hiter_datum.data);
	}
      chash_free(hash);
    }

  if ((hash = processing->actions))
    {
      for (hiter = chash_begin(hash); hiter; hiter = chash_next(hash, hiter))
	{
	  chash_value(hiter, &hiter_datum);
	  if (hiter_datum.data)
	    a_Processing_action_free((XePProcessingAction *)hiter_datum.data);
	}
      chash_free(hash);
    }
  x_free(processing);
}

XePProcessingCondition *
a_Processing_condition_new(void)
{
  XePProcessingCondition *condition;

  condition = x_alloc(1, XePProcessingCondition);

  if (condition == NULL)
    A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));

  return condition;
}

void
a_Processing_condition_free(XePProcessingCondition *condition)
{
  if (condition == NULL)
    return;

  A_LOGGING_DEBUG("freeing condition '%s'", condition->id);

  x_free(condition->id);
  x_free(condition->expression);

  x_free(condition);
}

int
a_Processing_condition_add(XePProcessing *processing, 
			   XePProcessingCondition *condition)
{
  chashdatum key;
  chashdatum data;

  if (!processing || !condition)
    {
      A_LOGGING_SYSTEM_ERROR("processing or condition NULL.");
       return -1;
    }

  A_LOGGING_DEBUG("adding condition '%s'", condition->id);
  
  key.data = condition->id;
  key.len  = strlen(condition->id);

  if (chash_get(processing->conditions, &key, &data) == 0)
    {
      A_LOGGING_ERROR("A condition with id='%s' is already here. "
		      "New condition ignored.", condition->id);
      goto err;
    }

  data.data = (char *) condition;
  data.len  = sizeof(condition);

  if (chash_set(processing->conditions, &key, &data, NULL))
    {
      A_LOGGING_SYSTEM_ERROR("chash_set error. Condition '%s' ignored.",
			     condition->id);
      goto err;
    }

  return 0;

 err:
  a_Processing_condition_free(condition);
  return -1;
}

static void *
Processing_condition_begin_iter(XePProcessing *processing)
{
  return chash_begin(processing->conditions);
}

static void *
Processing_condition_next_iter(XePProcessing *processing, void *iter)
{
  return chash_next(processing->conditions, (chashiter *) iter);
}

static XePProcessingCondition *
Processing_condition_get_from_iter(XePProcessing *processing, void *iter)
{
  chashdatum result;
  chash_value(iter, &result);
  return (XePProcessingCondition *) result.data;
}

XePProcessingAction *
a_Processing_action_new(void)
{
  XePProcessingAction *action;

  action = x_alloc(1, XePProcessingAction);

  if (action == NULL)
    A_LOGGING_SYSTEM_ERROR("calloc: %s", strerror(errno));

  return action;
}

void
a_Processing_action_free(XePProcessingAction *action)
{
  if (action == NULL)
    return;

  A_LOGGING_DEBUG("freeing action '%s'", action->id);

  free(action->id);

  free(action->command);

  free(action);
}

int
a_Processing_action_add(XePProcessing *processing, XePProcessingAction *action)
{
  chashdatum key;
  chashdatum data;

  if (!processing || !action)
    {
      A_LOGGING_SYSTEM_ERROR("processing or action NULL.");
       return -1;
    }

  A_LOGGING_DEBUG("adding action %s", action->id);

  key.data = action->id;
  key.len  = strlen(action->id);

  if (chash_get(processing->actions, &key, &data) == 0)
    {
      A_LOGGING_ERROR("An action with id='%s' is already here. "
		      "New action ignored.", action->id);
      goto err;
    }

  data.data = (char *) action;
  data.len  = sizeof(action);

  if (chash_set(processing->actions, &key, &data, NULL))
    {
      A_LOGGING_SYSTEM_ERROR("chash_set error. Action '%s' ignored.",
		      action->id);
      goto err;
    }

  return 0;

 err:
  a_Processing_action_free(action);
  return -1;

}

static void *
Processing_action_begin_iter(XePProcessing *processing)
{
  return chash_begin(processing->actions);
}

static void *
Processing_action_next_iter(XePProcessing *processing, void *iter)
{
  return chash_next(processing->actions, (chashiter *) iter);
}

static XePProcessingAction *
Processing_action_get_from_iter(XePProcessing *processing, void *iter)
{
  chashdatum result;
  chash_value(iter, &result);
  return (XePProcessingAction *) result.data;
}

XePProcessingRule *
a_Processing_rule_new(void)
{
  XePProcessingRule *rule;

  rule = x_alloc(1, XePProcessingRule);

  if (rule == NULL)
    A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));

  return rule;
}

void
a_Processing_rule_free(XePProcessingRule *rule)
{
  if (rule == NULL)
    return;
  
  A_LOGGING_DEBUG("freeing rule '%s'", rule->id);

  x_free(rule->id);
  x_free(rule->condition_ref);
  x_free(rule->action_ref);  

  x_free(rule);
}

XePProcessingRef *
a_Processing_ref_new(void)
{
  XePProcessingRef *ref;

  ref = x_alloc(1, XePProcessingRef);
 
  if (ref == NULL)
    A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
  else
    {
      ref->ref = NULL;
      ref->enabled = 1;
      ref->periodicity = 0;
    }
  
  return ref;
}

XePProcessingRef *
a_Processing_ref_new_copy(XePProcessingRef *ref)
{
  XePProcessingRef *ref_copy;
  
  ref_copy = a_Processing_ref_new();

  if (ref_copy)
    {
      ref_copy->ref         = X_STRDUP(ref->ref);
      ref_copy->enabled     = ref->enabled;
      ref_copy->periodicity = ref->periodicity;
    }

  return ref_copy;  
}

void
a_Processing_ref_free(XePProcessingRef *ref)
{
  if (ref == NULL)
    return;
  
  x_free(ref->ref);
  x_free(ref);
}

clist *
a_Processing_ref_list_new(void)
{
  return clist_new();
}

void
a_Processing_ref_list_free(clist *refs)
{
  clistiter *iter;

  if (refs == NULL)
    return;

  for (iter = clist_begin(refs); iter; iter = iter->next)
    a_Processing_ref_free((XePProcessingRef *)iter->data);

  clist_free(refs);
}

int
a_Processing_rule_add(XePProcessing *processing, XePProcessingRule *rule)
{
  chashdatum key;
  chashdatum data;

  if (!processing || !rule)
    {
      A_LOGGING_SYSTEM_ERROR("processing or rule NULL.");
       return -1;
    }

  A_LOGGING_DEBUG("adding rule '%s'", rule->id);

  key.data = rule->id;
  key.len  = strlen(rule->id);

  if (chash_get(processing->rules, &key, &data) == 0)
    {
      A_LOGGING_ERROR("A rule with id='%s' is already here. "
		      "New rule ignored.", rule->id);
      goto err;
    }

  data.data = (char *) rule;
  data.len  = sizeof(rule);

  if (chash_set(processing->rules, &key, &data, NULL))
    {
      A_LOGGING_SYSTEM_ERROR("chash_set error. Rule '%s' ignored.",
		      rule->id);
      goto err;
    }

  return 0;

 err:
  a_Processing_rule_free(rule);
  return -1;
}

static void *
Processing_rule_begin_iter(XePProcessing *processing)
{
  return chash_begin(processing->rules);
}

static void *
Processing_rule_next_iter(XePProcessing *processing, void *iter)
{
  return chash_next(processing->rules, (chashiter *) iter);
}

static XePProcessingRule *
Processing_rule_get_from_iter(XePProcessing *processing, void *iter)
{
  chashdatum result;
  chash_value(iter, &result);
  return (XePProcessingRule *) result.data;
}

int
a_Processing_create(cntree *ntree, XePProcessing **processing, int op)
{
  int r;
  LXMLelement *element;
  LXMLelement_callbacks element_cb[] = 
    {
      {"conditions", Processing_conditions_cb, NULL},
      {"actions",    Processing_actions_cb,    NULL},
      {"rules",      Processing_rules_cb,      NULL},
      {NULL, NULL, NULL}
    };
  LXMLelement_callbacks *a_element_cb;

  element = lxml_tree_get_element(ntree);

  if (op == LXML_Read)
    {
      *processing = a_Processing_new();
      if (*processing == NULL)
	return -1;
    }
  else if (*processing == NULL)
    return 0;
    

  for (a_element_cb = element_cb; 
       a_element_cb && a_element_cb->element_name;
       a_element_cb++)
    a_element_cb->data = (void **) processing;

  A_LOGGING_DEBUG("Parsing <processing>...");
  r = lxml_tree_convert(ntree, element_cb, op);
  A_LOGGING_DEBUG("<processing> parsed");

  if (op == LXML_Read && r == 0)
    {
      if (a_Processing_update_rules_links(*processing))
	{
	  A_LOGGING_ERROR("Processing data discarded");
	  a_Processing_free(*processing);
	  *processing = NULL;
	  return -1;
	}
    }
  return r;
}

static int
Processing_conditions_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessing **processing = (XePProcessing **) data;
  LXMLelement_callbacks element_cb[] = 
    {
      {"condition", Processing_condition_cb, NULL},
      {NULL, NULL, NULL}
    };

  element = lxml_tree_get_element(atree);

  element_cb[0].data = (void **) processing;

  A_LOGGING_DEBUG("Parsing <condition>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<condition> parsed");

  return r;
}

static int
Processing_actions_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessing **processing = (XePProcessing **) data;
  LXMLelement_callbacks element_cb[] = 
    {
      {"action", Processing_action_cb, NULL},
      {NULL, NULL, NULL}
    };

  element = lxml_tree_get_element(atree);

  element_cb[0].data = (void **) processing;

  A_LOGGING_DEBUG("Parsing <action>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<action> parsed");

  return r;
}

static int
Processing_rules_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessing **processing = (XePProcessing **) data;
  LXMLelement_callbacks element_cb[] = 
    {
      {"rule", Processing_rule_cb, NULL},
      {NULL, NULL, NULL}
    };

  element = lxml_tree_get_element(atree);

  element_cb[0].data = (void **) processing;

  A_LOGGING_DEBUG("Parsing <rule>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<rule> parsed");

  return r;
}

static int
Processing_condition_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessing *processing = *((XePProcessing **) data);
  XePProcessingCondition *condition;
  char *id = NULL;
  char *expression = NULL;
  LXMLstore store[] = 
    {
      {"id",         LXML_String, NULL, LXML_Unique},
      {"expression", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if ((NULL == *data_iter) && 
	  (NULL == (*data_iter = Processing_condition_begin_iter(processing))))
	return 0;

      condition = Processing_condition_get_from_iter(processing, *data_iter);

      *data_iter = Processing_condition_next_iter(processing, *data_iter);

      id = condition->id;
      expression = condition->expression;
    }

  store[0].dest = (void **) &id;
  store[1].dest = (void **) &expression;

  r= lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("Condition syntax error. Ignored.");
	  goto err;
	}

      if (!(id && expression))
	{
	  if (id)
	    A_LOGGING_ERROR("Condition '%s' does not have all attributes. "
			    "Ignored.", id);
	  else
	    A_LOGGING_ERROR("Condition is missing the id attribute.");
	  return -1;
	}

      condition = a_Processing_condition_new();

      if (condition == NULL)
	goto err;
      
      condition->id = id;
      condition->expression = expression;
      
      if (a_Processing_condition_add(processing, condition))
	{
	  a_Processing_condition_free(condition);
	  return -1;
	}
    }

  return 0;

 err:
  if (id)
    free(id);
  if (expression)
    free(expression);
  return -1;
}

static int
Processing_action_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessing *processing = *((XePProcessing **) data);
  XePProcessingAction *action;
  char *id = NULL;
  char *command = NULL;
  LXMLstore store[] = 
    {
      {"id",         LXML_String, NULL, LXML_Unique},
      {"expression", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };
  
  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if ((NULL == *data_iter) &&
	  (NULL == (*data_iter = Processing_action_begin_iter(processing))))
	return 0;
	  
      action = Processing_action_get_from_iter(processing, *data_iter);

      *data_iter = Processing_action_next_iter(processing, *data_iter);

      id = action->id;
      command = action->command;
    }

  store[0].dest = (void **) &id;
  store[1].dest = (void **) &command;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("Action syntax error. Ignored.");
	  goto err;
	}
      
      if (!(id && command))
	{
	  if (id)
	    A_LOGGING_ERROR("Action '%s' does not have all attributes."
			    " Ignored.", id);
	  else
	    A_LOGGING_ERROR("Action is missing the id attribute.");
	  return -1;
	}

      action = a_Processing_action_new();
      if (action == NULL)
	goto err;
      
      action->id = id;
      action->command = command;
      
      if (a_Processing_action_add(processing, action))
	{
	  a_Processing_action_free(action);
	  return -1;
	}
    }

  return r;

 err:
  x_free(id);
  x_free(command);
  return -1;
}

static int
Processing_rule_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessing *processing = *((XePProcessing **) data);
  XePProcessingRule *rule;
  char *id = NULL;
  char *condition_ref = NULL;
  char *action_ref = NULL;
  LXMLstore store[] = 
    {
      {"id",            LXML_String, NULL, LXML_Unique},
      {"condition_ref", LXML_String, NULL, LXML_Unique},
      {"action_ref",    LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };
  
  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if ((NULL == *data_iter) &&
	  (NULL == (*data_iter = Processing_rule_begin_iter(processing))))
	return 0;

      rule = Processing_rule_get_from_iter(processing, *data_iter);

      *data_iter = Processing_rule_next_iter(processing, *data_iter);

      id = rule->id;
      condition_ref = (rule->condition ? 
		       rule->condition->id : 
		       rule->condition_ref);
      action_ref = (rule->action ?
		    rule->action->id :
		    rule->action_ref);
    }
  
  store[0].dest = (void **) &id;
  store[1].dest = (void **) &condition_ref;
  store[2].dest = (void **) &action_ref;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("Rule syntax error. Ignored.");
	  goto err;
	}
      
      if (!id || !(condition_ref || action_ref))
	{
	  if (id)
	    A_LOGGING_ERROR("Rule '%s' does not have necessary attributes. "
			    "Ignored.", id);
	  else
	    A_LOGGING_ERROR("Rule is missing the id attribute");
	  goto err;
	}
      
      rule = a_Processing_rule_new();
      if (rule == NULL)
	goto err;
      
      rule->id = id;
      rule->condition_ref = condition_ref;
      rule->action_ref = action_ref;

      a_Processing_rule_add(processing, rule);
    }

  return r;

 err:
  if (id)
    free(id);
  if (condition_ref)
    free(condition_ref);
  if (action_ref)
    free(action_ref);
  return -1;
}

int
a_Processing_update_rules_links(XePProcessing *processing)
{
  chashdatum key;
  chashdatum datum;
  chashiter  *hiter;
  chash *rules;
  chash *conditions;
  chash *actions;    
  XePProcessingRule *rule;

  if (processing == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("Attempt to update rules links in "
			     "void processing list");
      return -1;
    }

  rules      = processing->rules;
  conditions = processing->conditions;
  actions    = processing->actions;

  if (rules == NULL)
    return 0;

  for (hiter = chash_begin(rules); hiter; hiter = chash_next(rules, hiter))
    {
      chash_value(hiter, &datum);
      rule = (XePProcessingRule *) datum.data;
      rule->condition = NULL;
      rule->action    = NULL;
      A_LOGGING_DEBUG("Updating rule '%s' links", rule->id);
      if (rule->condition_ref)
	{
	  key.data = rule->condition_ref;
	  key.len  = strlen(key.data);
	  if (conditions && chash_get(conditions, &key, &datum) == 0)
	    {
	      rule->condition = (XePProcessingCondition *) datum.data;
	      free(rule->condition_ref);
	      rule->condition_ref = NULL;
	      A_LOGGING_DEBUG("Rule '%s' linked to condition '%s'",
			     rule->id, rule->condition->id);
	    }
	  else
	    {
	      A_LOGGING_ERROR("Rule '%s' references unknown condition '%s'",
			      rule->id, rule->condition_ref);
	      return -1;
	    }
	  if (rule->action_ref)
	    {
	      key.data = rule->action_ref;
	      key.len  = strlen(key.data);
	      if (actions && (chash_get(actions, &key, &datum) == 0))
		{
		  rule->action = (XePProcessingAction *) datum.data;
		  free(rule->action_ref);
		  rule->action_ref = NULL;
		  A_LOGGING_DEBUG("Rule '%s' linked to action '%s'",
				 rule->id, rule->action->id);
		}
	      else
		{
		  A_LOGGING_ERROR("Rule '%s' references unknown action '%s'",
				  rule->id, rule->action_ref);
		  return -1;
		}
	    }
	}
    }
  return 0;
}

void *
Processing_get_from_ref(chash *hash, char *ref)
{
  chashdatum key;
  chashdatum datum;

  if (!hash || !ref)
    return NULL;

  key.data = ref;
  key.len  = strlen(key.data);

  if (chash_get(hash, &key, &datum))
    return NULL;

  return datum.data;
}

XePProcessingRule *
a_Processing_rule_get_from_ref(chash *rules, char *ref)
{
  return (XePProcessingRule *) Processing_get_from_ref(rules, ref);
}

XePProcessingCondition *
a_Processing_condition_get_from_ref(chash *rules, char *ref)
{
  return (XePProcessingCondition *) Processing_get_from_ref(rules, ref);
}

int
a_Processing_ref_list_copy(clist *source, clist *target)
{
  clistiter *iter;
  XePProcessingRef *ref;
  XePProcessingRef *ref_copy;
  int r = 0;

  /* TODO: this is actually an append instead of a copy */
  for (iter = clist_begin(source); iter && r == 0; iter = iter->next)
    {
      ref = (XePProcessingRef *) iter->data;
      ref_copy = a_Processing_ref_new_copy(ref);
      if (ref_copy)
	{
	  if (clist_append(target, ref_copy))
	    {
	      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
	      a_Processing_ref_free(ref_copy);
	      r = -1;
	    }
	}
      else
	{
	  r = -1;
	}
    }
  return r;
}
